#!/usr/bin/env python3
"""A very simple ocm implementation"""

__version__ = "0.0.1"

import argparse
import sys
import yaml
import uuid
import json
from http.server import CGIHTTPRequestHandler, HTTPServer
from os import environ as env
from os import path, makedirs, stat
from urllib.parse import urlparse

import pwd
import requests


class OCMHandler(CGIHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cgi_directories = ["/ocm"]

    def do_GET(self):
        if self.path in ["/.well-known/ocm", "/ocm-provider"]:
            self.send_response(200)
            self.end_headers()
            self.wfile.write(self.discovery())
        elif self.path == "/ocm/webdav":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"Hello World!")
        elif self.path == "/ocm/webapp":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"Hello World!")
        elif self.path == "/ocm/datatx":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"Hello World!")
        else:
            self.send_response(404)
            self.end_headers()
            self.wfile.write(b"Not Found")

    def do_POST(self):
        if self.path == "/ocm/shares":
            length = int(self.headers.get('content-length'))
            payload = json.loads(self.rfile.read(length))
            code, reply = self.validate_shares(payload)
            self.send_response(code)
            self.end_headers()
            self.wfile.write(json.dumps(reply).encode("utf-8"))
        elif self.path == "/ocm/webdav":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"{}")
        elif self.path == "/ocm/webapp":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"{}")
        elif self.path == "/ocm/datatx":
            self.send_response(200)
            self.end_headers
        else:
            self.send_response(404)
            self.end_headers()
            self.wfile.write(b"Not Found")

    def validate_shares(self, share) -> tuple[int, dict]:
        # spec: https://cs3org.github.io/OCM-API/docs.html?branch=develop&repo=OCM-API&user=cs3org#/paths/~1shares/post        return (201, {})
        share_with = share.get("shareWith")
        displayname = find_displayname(share_with)
        if displayname is None:
            return (400, {})
        reply = {
            "recipientDisplayName": displayname
        }
        return (201, reply)

    def discovery(self):
        # spec: https://cs3org.github.io/OCM-API/docs.html?branch=develop&repo=OCM-API&user=cs3org#/paths/~1.well-known~1ocm/get
        return json.dumps({
            "enabled": True,
            "apiVersion": "1.1.0",
            "endPoint": "http://127.0.0.1:8080/ocm",
            "provider": "ocm-cli",
            "resourceTypes": [
                {
                    "name": "file",
                    "shareTypes": [
                        "user"
                    ],
                    "protocols": {
                        "webdav": "http://127.0.0.1:8080/ocm/webdav",
                        "datatx": "http://127.0.0.1:8080/ocm/datatx"
                    }
                },
                {
                    "name": "folder",
                    "shareTypes": [
                        "user"
                    ],
                    "protocols": {
                        "webdav": "http://127.0.0.1:8080/ocm/webdav",
                        "datatx": "http://127.0.0.1:8080/ocm/datatx"
                    }
                }
            ],
            "capabilities": [
                "/invite-accepted"
            ]
        }).encode('utf-8')


def find_displayname(user: str) -> str:
    return pwd.getpwnam(user).pw_gecos.strip(',')


class OCM:
    port: int
    host: str
    server: HTTPServer
    dir: str
    config_dir: str
    config: dict
    config_file: str
    user: str
    fullname: str

    def __init__(self, dir: str = env["PWD"]):
        self.user = env["USER"]
        self.fullname = find_displayname(self.user)
        self.dir = dir
        self.config_dir = path.join(env["HOME"], ".config/ocm")
        self.config_file = path.join(self.config_dir, "config.yaml")
        if not path.isdir(self.config_dir):
            makedirs(self.config_dir)
        if not path.isfile(self.config_file):
            with open(self.config_file, "w") as f:
                config = {self.user: {}}
                f.write(yaml.dump(config))
        self.read_config()

    def read_config(self):
        with open(self.config_file, "r") as f:
            self.config = yaml.load(f, Loader=yaml.FullLoader)

    def write_config(self):
        with open(self.config_file, "w") as f:
            f.write(yaml.dump(self.config))

    def get_share_payload(self, provider_id: str) -> dict:
        name = self.config[self.user][provider_id]["name"],
        name = name[0]

        owner = pwd.getpwuid(stat(name).st_uid).pw_name
        return {
            "shareWith": self.config[self.user][provider_id]["shareWith"],
            "name": name,
            "providerId": provider_id,
            "owner": owner,
            "sender": self.user,
            "senderDisplayName": self.fullname,
            "shareType": self.config[self.user][provider_id]["shareType"],
            "resourceType": self.config[self.user][provider_id]["resourceType"],
            "protocol": {
                "name": "webdav",
                "webdav": {
                    "sharedSecret":
                        self.config[self.user][provider_id]["token"],
                    "permissions": self.config[self.user][provider_id]["permissions"],
                    "uri": f'/{self.user}/{provider_id}'
                }
            },
        }

    def main(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(dest="command")
        receive_parser = subparsers.add_parser("receive")
        receive_parser.add_argument("--port", type=int, default=8080)
        receive_parser.add_argument("--host", type=str, default="localhost")
        receive_parser.add_argument("--directory", type=str, default=self.dir)

        share_parser = subparsers.add_parser("share")
        share_group = share_parser.add_mutually_exclusive_group(required=True)
        share_group.add_argument("--file", type=str)
        share_group.add_argument("--directory", type=str)
        share_parser.add_argument("--to", type=str, required=True)
        args = parser.parse_args()

        if args.command == "receive":
            self.port = args.port
            self.host = args.host
            self.dir = args.directory
            self.server = HTTPServer((self.host, self.port), OCMHandler)
            self.receive()
        elif args.command == "share":
            self.share(args.file, args.directory, args.to)
        else:
            parser.print_help()

    def receive(self):
        self.server.serve_forever()

    def create_share(
            self,
            receiver,
            resource_type,
            name,
            share_endpoint
    ) -> None:
        provider_id = uuid.uuid4().hex
        self.config[self.user][provider_id] = {
            "name": name,
            "permissions": ["read"],
            "resourceType": resource_type,
            "shareType": "user",
            "shareWith": receiver,
            "token": uuid.uuid4().hex,
        }
        self.write_config()
        payload = self.get_share_payload(provider_id)

        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        share_req = requests.post(
            share_endpoint,
            headers=headers,
            json=payload)
        if share_req.status_code != 201:
            print("Error: " + str(share_req.status_code))
            print(share_req.text)
            sys.exit(1)
        share_data = share_req.json()
        print(share_data)

    def share(self, file, directory, to):
        uparse = urlparse(to)
        sharee = uparse.username
        print("Sharing with " + sharee)
        discovery_endpoint = uparse.scheme + "://" + \
            uparse.hostname + f":{uparse.port}/.well-known/ocm"
        req = requests.get(discovery_endpoint)
        if req.status_code != 200:
            print("Could not find ocm-provider at " + discovery_endpoint)
            sys.exit(1)
        data = req.json()
        share_endpoint = data['endPoint'] + "/shares"
        print(share_endpoint)
        enabled = data['enabled']
        if not enabled:
            print("OCM is not enabled on this server")
            sys.exit(1)
        resource_types = data['resourceTypes']
        file_or_folder = file
        resource_type = "file"
        if directory:
            file_or_folder = directory
            resource_type = "folder"

        type_supported = False
        for type in resource_types:
            if type['name'] == resource_type:
                type_supported = True
                break
        if not type_supported:
            print("Resource type " + resource_type + " is not supported")
            sys.exit(1)

        self.create_share(
            sharee, resource_type, file_or_folder, share_endpoint)


if __name__ == "__main__":
    try:
        ocm = OCM()
        ocm.main()
    except KeyboardInterrupt:
        print("Exiting...")
        sys.exit(0)
